from typing import Awaitable

import asyncio
from urllib.parse import urlparse, parse_qsl

import aioredis
from aioredis.pubsub import Receiver
from aioredis.abc import AbcChannel

from radical.log import logger
from radical import exceptions

QUEUE_PREFIX = 'radical:'


class RedisTransport(object):
    def __init__(self, transport_url, queue_name, loop):
        self.transport_url = transport_url
        self.queue_name = queue_name
        self.loop = loop
        self.conn = None
        config = dict(parse_qsl(urlparse(transport_url).query))
        self.request_timeout = int(config.get('request_timeout', 10))

    async def start(self):
        self.pool = await aioredis.create_redis_pool(
            self.transport_url,
            loop=self.loop,
            minsize=4  # At lest 2: one for BLPOP, other for PUBLISH
        )
        logger.debug('Redis transport started')

    async def stop(self):
        self.pool.close()
        await self.pool.wait_closed()

    async def get_next_request(self):
        source_name = QUEUE_PREFIX + self.queue_name
        try:
            result = await self.pool.execute(
                'blpop',
                source_name,
                1
            )
            if result is not None:
                logger.debug('BLPOP %s', source_name)
                return result[1]
        except Exception as error:
            logger.error(f'ERROR: {repr(error)}, retrying in 1 second')
            await asyncio.sleep(1)

    async def reply_to(self, request_id, message):
        source_name = QUEUE_PREFIX + request_id
        logger.debug('PUBLISH %s', source_name)
        await self.pool.execute(
            'publish',
            source_name,
            message
        )

    async def send_to(self, queue_name, message):
        source_name = QUEUE_PREFIX + queue_name
        logger.debug('LPUSH %s', source_name)
        await self.pool.execute('lpush', source_name, message)

    async def get_response(self, request_id) -> Awaitable:
        channel_name = QUEUE_PREFIX + request_id
        logger.debug('SUBSCRIBE %s', channel_name)
        channel = (await self.pool.subscribe(channel_name))[0]
        async def get_message():
            wait_coro = asyncio.ensure_future(channel.wait_message())
            timeout_coro = asyncio.ensure_future(asyncio.sleep(self.request_timeout))
            done = set()
            try:
                done, running = await asyncio.wait([
                    wait_coro, timeout_coro
                ], return_when=asyncio.FIRST_COMPLETED)
            except Exception as error:  # pragma: no cover
                logger.error('Error while waiting for message:', error)
                raise exceptions.RadicalException(f'Error while waiting for message: {str(error)}')
            else:
                if timeout_coro in done:
                    raise exceptions.TimeoutException('Timeout while waiting for response.')
                message = await channel.get()
                return message
            finally:
                timeout_coro.cancel()
                try:
                    await self.pool.unsubscribe(channel_name)
                except Exception as error:  # pragma: no cover
                    raise exceptions.TransportException(str(error))
                channel.close()
        return get_message()



import sys
import asyncio
import logging
from unittest import TestCase

import aioredis
import pytest

from radical.worker import Worker
from radical.client import Client
from radical.decorators import method
from radical.exceptions import RadicalException, TimeoutException

pytestmark = pytest.mark.asyncio

logging.basicConfig(format='%(asctime)s %(levelname)8s %(message)s', level=logging.DEBUG)


def create_pair(event_loop, request_timeout=5, **kwargs):
    worker = Worker(f'redis://redis:6379/0', 'test', loop=event_loop, **kwargs)
    client = Client(f'redis://redis:6379/0?request_timeout={request_timeout}', 'test', loop=event_loop, **kwargs)
    worker.register_method(
        'test.add',
        lambda a, b: a + b
    )

    async def wait(delay, result=42):
        return await asyncio.sleep(delay, result=result)

    worker.register_method(
        'test.wait',
        wait
    )

    return worker, client


async def test_call(event_loop):
    worker, client = create_pair(event_loop)
    await worker.start()
    await client.start()
    assert (await client.call('test', 'test.add', 1300, 37)) is None
    await worker.stop()
    await client.stop()


async def test_call_wait(event_loop):
    worker, client = create_pair(event_loop)
    await worker.start()
    await client.start()
    assert (await client.call_wait('test', 'test.add', 1300, 37)) == 1337
    await worker.stop()
    await client.stop()


async def test_call_wait_async(event_loop):
    worker, client = create_pair(event_loop)
    await worker.start()
    await client.start()
    assert (await client.call_wait('test', 'test.wait', 1, result=42)) == 42
    await worker.stop()
    await client.stop()


async def test_call_wait_error(event_loop):
    worker, client = create_pair(event_loop)
    await worker.start()
    await client.start()
    raised = False
    try:
        await client.call_wait('test', 'test.add', 1, '2')
    except RadicalException:
        raised = True
    assert raised, 'RadicalException was not raised.'
    await worker.stop()
    await client.stop()


async def test_json_serializer(event_loop):
    worker, client = create_pair(
        event_loop,
        serializer='radical.serialization.json:JSONSerializer'
    )
    await worker.start()
    await client.start()
    assert (await client.call_wait('test', 'test.add', 1300, 37)) == 1337
    await worker.stop()
    await client.stop()


async def test_bad_request(event_loop):
    worker, _ = create_pair(event_loop)
    await worker.start()
    conn = await aioredis.create_redis('redis://redis:6379/0')
    await conn.execute('lpush', 'radical:test', 'dafuq!')
    conn.close()
    await conn.wait_closed()
    # assert (await client.call_wait('test', 'test.add', 1300, 37)) == 1337
    await worker.stop()


async def test_bad_request_json(event_loop):
    worker, _ = create_pair(
        event_loop,
        serializer='radical.serialization.json:JSONSerializer'
    )
    await worker.start()
    conn = await aioredis.create_redis('redis://redis:6379/0')
    await conn.execute('lpush', 'radical:test', 'dafuq!')
    conn.close()
    await conn.wait_closed()
    # assert (await client.call_wait('test', 'test.add', 1300, 37)) == 1337
    await worker.stop()


async def test_timeout(event_loop):
    worker, client = create_pair(event_loop, request_timeout=2)
    await worker.start()
    await client.start()
    try:
        assert (await client.call_wait('unexistent', 'foo.bar'))
    except TimeoutException:
        raised = True
    assert raised, 'TimeoutException was not raised.'
    await worker.stop()
    await client.stop()


async def test_inspect(event_loop):
    worker, client = create_pair(event_loop, request_timeout=2)
    await worker.start()
    await client.start()
    actual = await client.call_wait('test', '_inspect')
    expected = ['test.add', 'test.wait', '_inspect']
    assert set(actual) == set(expected)
    await worker.stop()
    await client.stop()


async def test_discovery(event_loop):
    class RadicalFakePackage(object):
        class FakeModule(object):
            @method
            def foo(self):  # pragma: no cover
                pass

            def bar(self):  # pragma: no cover
                pass

    try:
        sys.modules['RadicalFakePackage'] = RadicalFakePackage
        sys.modules['RadicalFakePackage.FakeModule'] = RadicalFakePackage.FakeModule
        worker, _ = create_pair(event_loop)
        worker.discover([
            'RadicalFakePackage.FakeModule'
        ])
        assert 'RadicalFakePackage.FakeModule.foo' in worker.methods
        assert 'RadicalFakePackage.FakeModule.bar' not in worker.methods
    finally:
        del sys.modules['RadicalFakePackage']
        del sys.modules['RadicalFakePackage.FakeModule']



import pickle

from radical.serialization.base import (
    ProtocolError, BaseSerializer, RadicalRequest, RadicalResponse
)


class PickleSerializer(BaseSerializer):
    def encode_request(self, request: RadicalRequest) -> bytes:
        return pickle.dumps(super().encode_request(request))

    def decode_response(self, data: bytes) -> RadicalResponse:
        return super().decode_response(pickle.loads(data))

    def decode_request(self, data: bytes) -> RadicalRequest:
        try:
            data = pickle.loads(data)
        except Exception as error:
            raise ProtocolError(str(error))
        return super().decode_request(data)

    def encode_response(self, data: RadicalResponse) -> bytes:
        return pickle.dumps(super().encode_response(data))



import asyncio

from django.conf import settings
from django.core.management.base import BaseCommand

from radical.worker import Worker


class Command(BaseCommand):
    def handle(self, *args, **kwargs):
        self.worker = Worker(
            settings.RADICAL_CONFIG['TRANSPORT_URL'],
            queue_name=settings.RADICAL_CONFIG.get('QUEUE_NAME'),
            transport=settings.RADICAL_CONFIG.get('TRANSPORT'),
            serializer=settings.RADICAL_CONFIG.get('SERIALIZER'),
            loop=asyncio.get_event_loop()
        )
        self.worker.discover(settings.RADICAL_CONFIG['MODULES'])
        for key, value in self.worker.methods.items():
            print(f'  - {key} -> {value}')
        self.worker.run_until_complete()



import asyncio
import importlib

from radical.serialization.base import (
    RadicalRequest, RadicalResponse, ProtocolError, BaseSerializer
)


class Peer(object):
    def __init__(
            self,
            transport_url: str,
            queue_name: str = None,
            transport: str = None,
            serializer: str = None,
            loop=None
    ):
        if queue_name is None:  # pragma: no cover
            queue_name = 'default'
        if transport is None:
            transport = 'radical.transports.redis:RedisTransport'
        if serializer is None:
            serializer = 'radical.serialization.pickle:PickleSerializer'
        if loop is None:  # pragma: no cover
            loop = asyncio.get_event_loop()
        self.loop = loop
        self.transport_url = transport_url
        self.transport = self._create_transport(transport, transport_url, queue_name)
        self.serializer = self._create_serializer(serializer)
        self.queue_name = queue_name

    def _create_transport(self, transport: str, transport_url: str, queue_name: str):
        module_name, _, class_name = transport.rpartition(':')
        transport_module = importlib.import_module(module_name)
        transport_class = getattr(transport_module, class_name)
        return transport_class(transport_url, queue_name, self.loop)

    def _create_serializer(self, serializer: str) -> BaseSerializer:
        module_name, _, class_name = serializer.rpartition(':')
        serializer_module = importlib.import_module(module_name)
        serializer_class = getattr(serializer_module, class_name)
        return serializer_class()




import uuid

from radical.peer import Peer
from radical.serialization.base import RadicalRequest, Signature
from radical.log import logger
from radical.exceptions import RadicalException


class Client(Peer):
    def __init__(
            self,
            transport_url: str,
            queue_name: str = None,
            transport: str = None,
            serializer: str = None,
            loop=None
    ):
        super().__init__(
            transport_url, queue_name, transport, serializer, loop
        )

    async def call(self, queue_name, method, *args, **kwargs):
        logger.debug('Calling %s from %s (nowait mode)', method, queue_name)
        radical_request = RadicalRequest(signature=Signature(
            method=method,
            args=args,
            kwargs=kwargs
        ), reply_to=None)
        data = self.serializer.encode_request(radical_request)
        await self.transport.send_to(queue_name, data)

    async def call_wait(self, queue_name, method, *args, **kwargs):
        logger.debug('Calling %s from %s (wait mode)', method, queue_name)
        message_id = uuid.uuid1().hex
        response_future = await self.transport.get_response(message_id)
        radical_request = RadicalRequest(signature=Signature(
            method=method,
            args=args,
            kwargs=kwargs
        ), reply_to=message_id)
        data = self.serializer.encode_request(radical_request)
        await self.transport.send_to(queue_name, data)
        response = await response_future
        # TODO: Add timeout
        radical_response = self.serializer.decode_response(response)
        if radical_response.error:
            raise RadicalException(radical_response.error)
        return radical_response.result

    async def start(self):
        await self.transport.start()

    async def stop(self):
        await self.transport.stop()



import sys
sys.path.append('.')

from typing import Union, Callable, Optional, NoReturn, Awaitable, List
import logging
import asyncio
import importlib
import signal
import traceback
import argparse
from urllib.parse import urlparse

from radical.peer import Peer
from radical.serialization.base import (
    RadicalRequest, RadicalResponse, NoRequest,
    ProtocolError, BaseSerializer
)
from radical.log import logger
from radical.meta import VERSION, FLAIR


class Worker(Peer):
    def __init__(
            self,
            transport_url: str,
            queue_name: str = None,
            transport: str = None,
            serializer: str = None,
            loop=None
    ):
        super().__init__(transport_url, queue_name, transport, serializer, loop)
        self.methods = {
            '_inspect': lambda: list(self.methods.keys())
        }
        self.futures = []
        self.terminated = True
        self._main = None

    def discover(self, arg: Union[list, tuple, str]) -> List[str]:
        methods = []
        if isinstance(arg, (list, tuple)):
            for module_name in arg:
                methods.extend(self.discover(module_name))
        else:
            module_name = arg
            module_obj = importlib.import_module(module_name)
            for key, value in vars(module_obj).items():
                if hasattr(value, '__radical__'):
                    cannonical_name = '.'.join([module_name, key])
                    self.register_method(cannonical_name, value)
                    methods.append(cannonical_name)
        return methods

    def register_method(self, cannonical_name: str, method: Callable) -> NoReturn:
        assert cannonical_name not in self.methods, \
            f'Cannonical name {cannonical_name} already registered.'
        self.methods[cannonical_name] = method

    async def start(self) -> Awaitable:
        urlinfo = urlparse(self.transport_url)
        sys.stderr.write(FLAIR % (
            f'Version   : {VERSION}',
            f'URL       : {urlinfo.scheme}://***@{urlinfo.hostname}',
            f'Queue     : {self.queue_name}',
            f'Transport : {self.transport.__class__.__name__}',
            f'Serializer: {self.serializer.__class__.__name__}',
        ))
        sys.stderr.write(f'\n')
        if len(self.methods):
            sys.stderr.write('Offering the following methods:\n\n')
            for method in self.methods.keys():
                sys.stderr.write(f'  - {method}\n')
        else:
            sys.stderr.write(f'  ! No methods found.\n')
        sys.stderr.write('\n')

        await self.transport.start()
        self._main = asyncio.ensure_future(self._run())
        return self._main

    async def _run(self) -> NoReturn:
        logger.info('Radical RPC server is ready.')
        self.terminated = False
        # asyncio.get_event_loop().create_task(self._accept())
        self.futures.append(asyncio.ensure_future(self._accept(), loop=self.loop))
        while len(self.futures):
            done, running = await asyncio.wait(
                self.futures,
                # [self._accept()] + self.request_futures,
                return_when=asyncio.FIRST_COMPLETED,
                loop=self.loop
            )
            while done:
                future = done.pop()
                self.futures.remove(future)
                obj = future.result()
                if isinstance(obj, RadicalRequest):
                    if not self.terminated:
                        self.futures.append(asyncio.ensure_future(self._accept(), loop=self.loop))
                    self.futures.append(
                        asyncio.ensure_future(self._process_request(obj), loop=self.loop)
                    )
                elif isinstance(obj, RadicalResponse):
                    self.futures.append(asyncio.ensure_future(self._process_response(obj), loop=self.loop))
                elif isinstance(obj, NoRequest):
                    if not self.terminated:
                        self.futures.append(asyncio.ensure_future(self._accept(), loop=self.loop))
        logger.info('Radical RPC server is terminating gracefully.')
        await self.transport.stop()

    async def _accept(self) -> RadicalRequest:
        request = await self.transport.get_next_request()
        if request is not None:
            try:
                radical_request = self.serializer.decode_request(request)
            except ProtocolError as error:
                logger.error(f'Deserialization failed: {error}')
            else:
                return radical_request
        return NoRequest()
            # asyncio.ensure_future(self._process_request(request))

    async def _process_request(self, radical_request: RadicalRequest) -> Optional[RadicalResponse]:
        logger.info(f'Received request {radical_request}')
        result, error = None, None
        try:
            signature = radical_request.signature
            method = self.methods[signature.method]
            if asyncio.iscoroutinefunction(method):
                result = await method(*signature.args, **signature.kwargs)
            else:
                result = method(*signature.args, **signature.kwargs)
        except Exception as method_error:
            # TODO: Include traceback
            error = str(method_error)
        if error:
            logger.error(f'WARNING: Error occured: {error}')
        return RadicalResponse(
            request=radical_request, result=result, error=error
        )

    async def _process_response(self, radical_response: RadicalResponse) -> NoReturn:
        if radical_response.request.reply_to:
            logger.info(f'Sending response {radical_response}')
            data = self.serializer.encode_response(radical_response)
            await self.transport.reply_to(
                radical_response.request.reply_to,
                data
            )
        else:
            logger.info(f'Discarding response {radical_response}')

    def _exception_handler(self, loop, data) -> NoReturn:
        logger.error(f'Exception in loop {loop}: {data}')

    def run_until_complete(self) -> NoReturn:
        self.loop.set_exception_handler(self._exception_handler)
        for sig in (signal.SIGINT, signal.SIGTERM):
            self.loop.add_signal_handler(
                sig,
                lambda: asyncio.ensure_future(self.stop(), loop=self.loop)
            )
        try:
            coro = self.loop.run_until_complete(self.start())
            self.loop.run_until_complete(coro)
        except Exception as error:
            logger.exception(str(error))
            asyncio.ensure_future(self.stop(), loop=self.loop)
            self.loop.stop()
        # pending = asyncio.Task.all_tasks(loop=self.loop)
        # self.loop.run_until_complete(asyncio.gather(*pending, loop=self.loop))

    async def stop(self):
        logger.info('Attempting graceful shutdown.')
        logger.info('Do NOT kill the process right now or you will lose data!')
        self.terminated = True
        logger.info('Waiting for all tasks to complete...')
        await self._main


def main():  # pragma: no cover
    parser = argparse.ArgumentParser()
    parser.add_argument('-u', '--url', default='redis://127.0.0.1:6379/0', dest='transport_url')
    parser.add_argument('-q', '--queue', default='default', dest='queue_name')
    parser.add_argument('-t', '--transport', default='radical.transports.redis:RedisTransport')
    parser.add_argument('-s', '--serializer', default='radical.serialization.pickle:PickleSerializer')
    parser.add_argument('-l', '--level', default='INFO', help='Logging level.')
    parser.add_argument('module', nargs='+', help='Module with methods. Can be specified multiple times.')
    args = dict(vars(parser.parse_args()))
    modules = args.pop('module')
    level = getattr(logging, args.pop('level').upper())
    logging.basicConfig(level=level, format='%(asctime)s [%(levelname)-8s] (%(module)s) %(message)s')
    worker = Worker(**args)
    methods = worker.discover(modules)
    worker.run_until_complete()


if __name__ == '__main__':  # pragma: no cover
    main()



#!/usr/bin/env python3.6

# coding=utf-8

from setuptools import setup
from radical.meta import VERSION

setup(
    name='radical-rpc',
    version=VERSION,
    description='Multi-transport RPC with asyncio & Django support.',
    author="Andrew Dunai",
    author_email='andrew@dun.ai',
    url='https://github.com/and3rson/radical',
    license='GPLv3',
    packages=[
        'radical',
        'radical.transports',
        'radical.management',
        'radical.management.commands',
        'radical.serialization',
        'radical.contrib',
        'radical.contrib.django',
    ],
    include_package_data=True,
    install_requires=['aioredis'],
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'radical=radical.worker:main'
        ]
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Framework :: Django',
        'Framework :: AsyncIO',
        'Topic :: Software Development :: Libraries',
    ],
    keywords='rpc,python2,python3,python,asyncio,aio,redis,django',
)



from collections import namedtuple


class ProtocolError(Exception):
    pass


RadicalRequest = namedtuple('RadicalRequest', ('signature', 'reply_to'))
Signature = namedtuple('Signature', ('method', 'args', 'kwargs'))
RadicalResponse = namedtuple('RadicalResponse', ('request', 'result', 'error'))
NoRequest = namedtuple('NoRequest', ())


class BaseSerializer(object):
    def encode_request(self, request: RadicalRequest) -> dict:
        return dict(
            method=request.signature.method,
            args=request.signature.args,
            kwargs=request.signature.kwargs,
            reply_to=request.reply_to
        )

    def decode_response(self, data: dict) -> RadicalResponse:
        return RadicalResponse(
            request=None,
            result=data.get('result'),
            error=data.get('error')
        )

    def decode_request(self, data: dict) -> RadicalRequest:
        if not isinstance(data, dict):  # pragma: no cover
            raise ProtocolError('request is not a dictionary.')
        if not isinstance(data.get('method'), str):  # pragma: no cover
            raise ProtocolError('".method" is missing or invalid.')
        if not isinstance(data.get('args', []), (list, tuple)):  # pragma: no cover
            raise ProtocolError('".args" is missing or invalid.')
        if not isinstance(data.get('kwargs', {}), dict):  # pragma: no cover
            raise ProtocolError('".kwargs" is missing or invalid.')
        return RadicalRequest(
            signature=Signature(
                method=data.get('method'),
                args=data.get('args', []),
                kwargs=data.get('kwargs', {})
            ),
            reply_to=data.get('reply_to', '')
        )

    def encode_response(self, response: RadicalResponse) -> dict:
        return dict(
            result=response.result,
            error=response.error
        )



import asyncio

from django.conf import settings

from radical.client import Client


def _create_session():
    loop = asyncio.new_event_loop()
    client = Client(
        settings.RADICAL_CONFIG['TRANSPORT_URL'],
        queue_name=settings.RADICAL_CONFIG.get('QUEUE_NAME'),
        transport=settings.RADICAL_CONFIG.get('TRANSPORT'),
        serializer=settings.RADICAL_CONFIG.get('SERIALIZER'),
        loop=loop
    )
    return loop, client


def _destroy_session(client):
    client.loop.run_until_complete(client.stop())
    client.loop.stop()
    client.loop.close()


def call(queue_name, method, *args, **kwargs):
    loop, client = _create_session()
    try:
        loop.run_until_complete(client.start())
        coroutine = client.call(queue_name, method, *args, **kwargs)
        loop.run_until_complete(coroutine)
    finally:
        _destroy_session(client)


def call_wait(queue_name, method, *args, **kwargs):
    loop, client = _create_session()
    try:
        loop.run_until_complete(client.start())
        coroutine = client.call_wait(queue_name, method, *args, **kwargs)
        return loop.run_until_complete(coroutine)
    finally:
        _destroy_session(client)
